export const textareaStyles = {
  variants: {
    outline: {
      borderRadius: "4px",
      borderColor: "gray.200",
      _hover: {
        borderColor: "gray.200",
      },
      _focus: {
        borderColor: "gray.200",
        boxShadow: "none",
      },
    },
  },
};
